"""Connection testing."""

import requests
import json
import pytest
import creopyson
from creopyson.exceptions import MissingKey, ErrorJsonDecode

from .fixtures import mk_creoson_post_None, mk_creoson_post_dict, \
    mk_creoson_post_sessionId

# @pytest.fixture(autouse=True)
# def no_requests(monkeypatch):
#     monkeypatch.delattr("requests.sessions.Session.request")


def test_connection_wether_params_exists():
    """Test wether connection vars are created and well generated."""
    c = creopyson.Client()
    assert c.server == "http://localhost:9056/creoson"
    assert c.sessionId == ''

    c = creopyson.Client(ip_adress="here", port=1234)
    assert c.server == "http://here:1234/creoson"


def test_connection_connect_succed(mk_creoson_post_sessionId):
    """Test when connection is ok.

    sessionId is created and retruened by creoson.

    """
    c = creopyson.Client()
    c.connect()
    assert c.sessionId == "123456"


def test_connection_connect_fails(monkeypatch):
    """Test when connection fails and exit."""
    class Mk_post():
        def __init__(self, *args, **kwargs):
            raise requests.exceptions.RequestException

    monkeypatch.setattr(requests, 'post', Mk_post)
    c = creopyson.Client()
    with pytest.raises(ConnectionError) as pytest_wrapped_e:
        c.connect()
    assert pytest_wrapped_e.type == ConnectionError



def test_connection_creoson_post_error_connection(monkeypatch):
    """Test _creoson_post connection server error."""
    class Mk_post():
        def __init__(self, *args, **kwargs):
            pass

        @property
        def status_code(self):
            return 500

    monkeypatch.setattr(requests, 'post', Mk_post)
    c = creopyson.Client()
    with pytest.raises(ConnectionError) as pytest_wrapped_e:
        c._creoson_post("function", "method", {})
    assert pytest_wrapped_e.value.args[0] == "Status code : 500"


def test_connection_creoson_post_json_error(monkeypatch):
    """Test _creoson_post whether result can't be decoded with json."""
    class Mk_post():
        def __init__(self, *args, **kwargs):
            pass

        def json(self):
            raise TypeError

        @property
        def status_code(self):
            return 200
    monkeypatch.setattr(requests, 'post', Mk_post)
    c = creopyson.Client()
    with pytest.raises(ErrorJsonDecode) as pytest_wrapped_e:
        c._creoson_post("function", "method", {})
    assert pytest_wrapped_e.value.args[0] ==\
        "Cannot decode JSON, creoson result invalid."


def test_connection_creoson_post_raise_RuntimeError(monkeypatch):
    """Test _creoson_post raise RuntimeError."""
    class Mk_post():
        def __init__(self, *args, **kwargs):
            pass

        def json(self):
            results = {
                "status": {
                    "error": True,
                    "message": "error message"
                }
            }
            return results

        @property
        def status_code(self):
            return 200

    monkeypatch.setattr(requests, 'post', Mk_post)
    c = creopyson.Client()
    with pytest.raises(RuntimeError) as pytest_wrapped_e:
        c._creoson_post("function", "method", {})
    assert pytest_wrapped_e.value.args[0] == "error message"


def test_connection_creoson_post_return_data(monkeypatch):
    """Test _creoson_post returning data."""
    class Mk_post():
        def __init__(self, *args, **kwargs):
            pass

        def json(self):
            results = {
                "status": {
                    "error": False,
                    "message": "error message"
                },
                "data": "creoson result"
            }
            return results

        @property
        def status_code(self):
            return 200

    monkeypatch.setattr(requests, 'post', Mk_post)
    c = creopyson.Client()
    result = c._creoson_post("function", "method", {})
    assert result == "creoson result"


def test_connection_creoson_post_return_None(monkeypatch):
    """Test _creoson_post returning None."""
    class Mk_post():
        def __init__(self, *args, **kwargs):
            pass

        def json(self):
            results = {
                "status": {
                    "error": False,
                    "message": "error message"
                }
            }
            return results

        @property
        def status_code(self):
            return 200

    monkeypatch.setattr(requests, 'post', Mk_post)
    c = creopyson.Client()
    result = c._creoson_post("function", "method", {})
    assert result is None


def test_connection_creoson_post_no_status(monkeypatch):
    """Test _creoson_post raise error if no `status` in return."""
    class Mk_post():
        def __init__(self, *args, **kwargs):
            pass

        def json(self):
            results = {"status": {}}
            return results

        @property
        def status_code(self):
            return 200

    monkeypatch.setattr(requests, 'post', Mk_post)
    c = creopyson.Client()
    with pytest.raises(MissingKey) as pytest_wrapped_e:
        c._creoson_post("function", "method", {})
    assert pytest_wrapped_e.value.args[0] ==\
        "Missing `error` in status' creoson's result."


def test_connection_creoson_post_no_error(monkeypatch):
    """Test _creoson_post raise error if no `error` in return."""
    class Mk_post():
        def __init__(self, *args, **kwargs):
            pass

        def json(self):
            results = {}
            return results

        @property
        def status_code(self):
            return 200

    monkeypatch.setattr(requests, 'post', Mk_post)
    c = creopyson.Client()
    with pytest.raises(MissingKey) as pytest_wrapped_e:
        c._creoson_post("function", "method", {})
    assert pytest_wrapped_e.value.args[0] ==\
        "Missing `status` in creoson result."


def test_connection_creoson_post_no_sessionId(monkeypatch):
    """Test _creoson_post raise error if no `sessionId` in return."""
    class Mk_post():
        def __init__(self, *args, **kwargs):
            pass

        def json(self):
            results = {
                "status": {
                    "error": False
                }
            }
            return results

        @property
        def status_code(self):
            return 200

    monkeypatch.setattr(requests, 'post', Mk_post)
    c = creopyson.Client()
    with pytest.raises(MissingKey) as pytest_wrapped_e:
        c._creoson_post("connection", "connect", {})
    assert pytest_wrapped_e.value.args[0] ==\
        "Missing `sessionId` in creoson result."


def test_connection_creoson_post_sessionId(monkeypatch):
    """Test _creoson_post return sessionId."""
    class Mk_post():
        def __init__(self, *args, **kwargs):
            pass

        def json(self):
            results = {
                "status": {
                    "error": False
                },
                "sessionId": 12345
            }
            return results

        @property
        def status_code(self):
            return 200

    monkeypatch.setattr(requests, 'post', Mk_post)
    c = creopyson.Client()
    result = c._creoson_post("connection", "connect", {})
    assert result == 12345


def test_connection_creoson_post_no_data(monkeypatch):
    """Test _creoson_post raise error if no data in return."""
    class Mk_post():
        def __init__(self, *args, **kwargs):
            pass

        def json(self):
            results = {
                "status": {
                    "error": False
                }
            }
            return results

        @property
        def status_code(self):
            return 200

    monkeypatch.setattr(requests, 'post', Mk_post)
    c = creopyson.Client()
    with pytest.raises(MissingKey) as pytest_wrapped_e:
        c._creoson_post("function", "method", {}, key_data="fakedata")
    assert pytest_wrapped_e.value.args[0] ==\
        "Missing `data` in creoson return"


def test_connection_creoson_post_no_datakey(monkeypatch):
    """Test _creoson_post raise error if no datakey in return."""
    class Mk_post():
        def __init__(self, *args, **kwargs):
            pass

        def json(self):
            results = {
                "status": {
                    "error": False
                },
                "data": {}
            }
            return results

        @property
        def status_code(self):
            return 200

    monkeypatch.setattr(requests, 'post', Mk_post)
    c = creopyson.Client()
    with pytest.raises(MissingKey) as pytest_wrapped_e:
        c._creoson_post("function", "method", {}, key_data="fakedata")
    assert pytest_wrapped_e.value.args[0] ==\
        "Missing `fakedata` in creoson result"


def test_connection_creoson_post_datakey(monkeypatch):
    """Test _creoson_post raise error if no datakey in return."""
    class Mk_post():
        def __init__(self, *args, **kwargs):
            pass

        def json(self):
            results = {
                "status": {
                    "error": False
                },
                "data": {
                    "fakedata": "fakevalue"
                }
            }
            return results

        @property
        def status_code(self):
            return 200

    monkeypatch.setattr(requests, 'post', Mk_post)
    c = creopyson.Client()
    result = c._creoson_post("function", "method", {}, key_data="fakedata")
    assert result == "fakevalue"


def test_connection_disconnect_ok(mk_creoson_post_None):
    """Test wether client is disconnected (empty sessionId)."""
    c = creopyson.Client()
    c.sessionId = "12345"
    c.disconnect()
    assert c.sessionId == ""


def test_connection_is_creo_running_yes(mk_creoson_post_dict):
    """Test wether creo is running OK."""
    c = creopyson.Client()
    result = c.is_creo_running()
    assert result


def test_connection_kill_creo_ok(mk_creoson_post_None):
    """Test no error returned from creoson."""
    c = creopyson.Client()
    result = c.kill_creo()
    assert result is None


def test_connection_start_creo_ok(mk_creoson_post_None):
    """Test no error returned from creoson."""
    c = creopyson.Client()
    result = c.start_creo("C:/path/to/nitro_proe_remote.bat")
    assert result is None


def test_connection_stop_creo_ok(mk_creoson_post_None):
    """Test no error returned from creoson."""
    c = creopyson.Client()
    result = c.stop_creo()
    assert result is None



"""Familytable testing."""
import creopyson
from .fixtures import mk_creoson_post_dict, mk_creoson_post_None, mk_getactivefile


def test_familytable_add_inst(mk_creoson_post_None, mk_getactivefile):
    """Test add_inst."""
    c = creopyson.Client()
    result = c.familytable_add_inst("instance", file_="file")
    assert result is None
    result = c.familytable_add_inst("instance")
    assert result is None


def test_familytable_create_inst(mk_creoson_post_dict, mk_getactivefile):
    """Test create_inst."""
    c = creopyson.Client()
    result = c.familytable_create_inst("instance", file_="file")
    assert isinstance(result, (str))
    result = c.familytable_create_inst("instance")
    assert isinstance(result, (str))


def test_familytable_delete_inst(mk_creoson_post_None, mk_getactivefile):
    """Test delete_inst."""
    c = creopyson.Client()
    result = c.familytable_delete_inst("instance", file_="file")
    assert result is None
    result = c.familytable_delete_inst("instance")
    assert result is None


def test_familytable_delete(mk_creoson_post_None, mk_getactivefile):
    """Test delete."""
    c = creopyson.Client()
    result = c.familytable_delete(file_="file")
    assert result is None
    result = c.familytable_delete()
    assert result is None


def test_familytable_exists(mk_creoson_post_dict, mk_getactivefile):
    """Test exists."""
    c = creopyson.Client()
    result = c.familytable_exists("instance", file_="file")
    assert isinstance(result, (bool))
    result = c.familytable_exists("instance")
    assert isinstance(result, (bool))


def test_familytable_get_cell(mk_creoson_post_dict, mk_getactivefile):
    """Test get_cell."""
    c = creopyson.Client()
    result = c.familytable_get_cell(
        "instance",
        12,
        file_="file"
    )
    assert isinstance(result, (dict))
    result = c.familytable_get_cell(
        "instance",
        12
    )
    assert isinstance(result, (dict))


def test_familytable_get_header(mk_creoson_post_dict, mk_getactivefile):
    """Test get_header."""
    c = creopyson.Client()
    result = c.familytable_get_header(file_="file")
    assert isinstance(result, (list))
    result = c.familytable_get_header()
    assert isinstance(result, (list))


def test_familytable_get_parents(mk_creoson_post_dict, mk_getactivefile):
    """Test get_parents."""
    c = creopyson.Client()
    result = c.familytable_get_parents(file_="file")
    assert isinstance(result, (list))
    result = c.familytable_get_parents()
    assert isinstance(result, (list))


def test_familytable_get_row(mk_creoson_post_dict, mk_getactivefile):
    """Test get_row."""
    c = creopyson.Client()
    result = c.familytable_get_row("instance", file_="file")
    assert isinstance(result, (list))
    result = c.familytable_get_row("instance")
    assert isinstance(result, (list))
    # it should have been a dict but colums already exist with list


def test_familytable_list(mk_creoson_post_dict, mk_getactivefile):
    """Test list."""
    c = creopyson.Client()
    result = c.familytable_list(file_="file", instance="instance")
    assert isinstance(result, (list))
    result = c.familytable_list()
    assert isinstance(result, (list))


def test_familytable_list_tree(mk_creoson_post_dict, mk_getactivefile):
    """Test list_tree."""
    c = creopyson.Client()
    result = c.familytable_list_tree(file_="file", erase=True)
    assert isinstance(result, (list))
    result = c.familytable_list_tree()
    assert isinstance(result, (list))


def test_familytable_replace(mk_creoson_post_None, mk_getactivefile):
    """Test replace."""
    c = creopyson.Client()
    result = c.familytable_replace(
        "cur_model",
        "new_inst",
        file_="file",
        cur_inst="cur_inst",
        path=[1, 2, 3]
    )
    assert result is None
    result = c.familytable_replace(
        "cur_model",
        "new_inst"
    )
    assert result is None


def test_familytable_set_cell(mk_creoson_post_None, mk_getactivefile):
    """Test set_cell."""
    c = creopyson.Client()
    result = c.familytable_set_cell(
        "instance",
        "col id",
        23,
        file_="file"
    )
    assert result is None
    result = c.familytable_set_cell(
        "instance",
        "col id",
        23
    )
    assert result is None



"""Fixtures for tests."""
import creopyson
import pytest


@pytest.fixture
def mk_creoson_post_dict(monkeypatch):
    """Mock _creoson_post return dict."""
    def fake_func(client, command, function, data=None, key_data=None):
        if key_data:
            result = {
                "active": True,
                "checked_out": True,
                "children": [],
                "columns": [],
                "contourlist": [],
                "dirname": "dirname",
                "dimlist": [],
                "dirlist": [],
                "drawing": "drawing",
                "exists": True,
                "errors": True,
                "featlist": [],
                "file": "file",
                "filelist": [],
                "files": [],
                "instances": [],
                "itemlist": [],
                "layers": [],
                "loaded": True,
                "material": "material",
                "materials": [],
                "name": "name",
                "num_sheets": 4,
                "paramlist": [],
                "parents": [],
                "relations": [],
                "running": True,
                "scale": 1.5,
                "size": "size",
                "sheet": 4,
                "surflist": [],
                "symbols": [],
                "transform": {"fake obj": 123},
                "units": "units",
                "values": [],
                "viewlist": [],
                "views": [],
                "workspace": "workspace",
                "workspaces": [],
            }
            return result[key_data]
        else:
            return {}
    monkeypatch.setattr(
        creopyson.connection.Client, '_creoson_post', fake_func)


@pytest.fixture
def mk_creoson_post_None(monkeypatch):
    """Mock _creoson_post return None."""
    def fake_func(client, command, function, data=None, key_data=None):
        return None
    monkeypatch.setattr(
        creopyson.connection.Client, '_creoson_post', fake_func)


@pytest.fixture
def mk_creoson_post_list(monkeypatch):
    """Mock _creoson_post return list."""
    def fake_func(client, command, function, data=None, key_data=None):
        return ['information']
    monkeypatch.setattr(
        creopyson.connection.Client, '_creoson_post', fake_func)


@pytest.fixture
def mk_creoson_post_sessionId(monkeypatch):
    """Mock _creoson_post return dict."""
    def fake_func(client, command, function, data=None, key_data=None):
        return "123456"
    monkeypatch.setattr(
        creopyson.connection.Client, '_creoson_post', fake_func)


@pytest.fixture
def mk_getactivefile(monkeypatch):
    """Mock file_get_active()."""
    def fake_func(client):
        return {
            "file": "file.prt"
        }
    monkeypatch.setattr(
        creopyson.connection.Client, "file_get_active", fake_func)



"""Server module."""

import requests
import json
import re


def pwd(client):
    """Return the creoson server's execution directory.

    Args:
        client (obj):
            creopyson Client.

    Raises:
        Warning: error message from creoson.

    Returns:
        (str): Full name of working directory.

    """
    request = {
        "sessionId": client.sessionId,
        "command": "server",
        "function": "pwd",
    }
    # ask `http://localhost:9056/server` vs `http://localhost:9056/creoson`
    server_adress = re.sub(r'creoson$', 'server',  client.server)
    r = requests.post(server_adress, data=json.dumps(request))
    json_result = r.json()
    status = json_result["status"]["error"]

    if not status:
        return json_result["data"]["dirname"]
    else:
        raise Warning(json_result["status"]["message"])



"""Creo module."""
from creopyson.exceptions import MissingKey


def cd(client, dirname):
    r"""Change Creo's working directory.

    You can use absolute path or relative:
    "C:\\My Workdir\\"
    "..\\Other_directory\\"

    Args:
        client (obj): creopyson Client.
        dirname (str): New directory name.

    Returns:
        (str): Name of new working directory.

    """
    data = {"dirname": dirname}
    return client._creoson_post("creo", "cd", data, "dirname")


def delete_files(client, filename=None, dirname=None):
    """Delete files from a directory working directory.

    Args:
        client (obj):
            creopyson Client.
        filename (str or list:str, optional):
            File name filter or list of file names.
            if blank all files will be deleted.
            (wildcards_allowed: True). Defaults to None.
        dirname (str, optional):
            Directory name. Defaults is Creo's current working directory.

    Returns:
        (list:str): List of deleted files.

    """
    data = {}
    if filename:
        if isinstance(filename, (list)):
            data["filenames"] = filename
        else:
            data["filename"] = str(filename)
    if dirname:
        data["dirname"] = dirname
    return client._creoson_post("creo", "delete_files", data, "filelist")


def get_config(client, name):
    """Get the value of a Creo config option.

    Args:
        client (obj): creopyson Client.
        name (str): Option name.

    Raises:
        Warning: error message from creoson.

    Returns:
        (list:str):
            List of option values (some options can have multiple values).

    """
    data = {"name": name}
    return client._creoson_post("creo", "get_config", data, "values")


def get_std_color(client, color_type):
    """Get one of Creo's standard colors.

    Args:
        client (obj):
            creopyson Client.
        color_type (str):
            Color type.
            Valid values: letter, highlight, drawing, background, half_tone,
            edge_highlight, dimmed, error, warning, sheetmetal, curve,
            presel_highlight, selected, secondary_selected, preview,
            secondary_preview, datum, quilt.

    Returns:
        (dict):
            red (int): Red value (0-255)
            green (int): Green value (0-255)
            blue (int): Blue value (0-255)

    """
    data = {"color_type": color_type}
    return client._creoson_post("creo", "get_std_color", data)


def list_dirs(client, dirname=None):
    """List subdirectories of Creo's current working directory.

    Args:
        client (obj):
            creoopyson Client.
        dirname (str, optional):
            Directory name filter (wildcards_allowed: True).
            Defaults: All subdirectories are listed.

    Returns:
        (list:str): List of subdirectories

    """
    data = {"dirname": "*"}
    if dirname:
        data["dirname"] = dirname
    try:
        result = client._creoson_post("creo", "list_dirs", data, "dirlist")
    except MissingKey:
        result = []
    return result


def list_files(client, filename=None):
    """List files in Creo's current working directory.

    Args:
        client (obj):
            creopyson Client.
        filename (str, optional):
            File name filter (wildcards_allowed: True).
            Defaults: all files are listed.

    Returns:
        (list:int): List of files.

    """
    data = {"filename": "*"}
    if filename:
        data["filename"] = filename
    return client._creoson_post("creo", "list_files", data, "filelist")


def mkdir(client, dirname):
    """Create a new directory.

    Args:
        client (obj): creopyson Client.
        dirname (str): New directory name.

    Returns:
        (str): Full name of new working directory.

    """
    data = {"dirname": dirname}
    return client._creoson_post("creo", "mkdir", data, "dirname")


def pwd(client):
    """Return Creo's current working directory.

    Args:
        client (obj): creopyson Client.

    Returns:
        (str): Full name of working directory.

    """
    return client._creoson_post("creo", "pwd", key_data="dirname")


def rmdir(client, dirname):
    """Delete a directory.

    Args:
        client (obj): creopyson Client.
        dirname (str): Directory name to delete.

    Returns:
        None.

    """
    data = {"dirname": dirname}
    return client._creoson_post("creo", "rmdir", data)


def set_config(client, name, value, ignore_errors=None):
    """Set a Creo config option.

    Args:
        client (obj):
            creopyson Client.
        name (str):
            Option name.
        value (str):
            New option value.
        ignore_errors (boolean, optional):
            Whether to ignore errors that might occur when setting the config
            option. Defaults is False.

    Returns:
        None.

    """
    data = {"name": name, "value": value, "ignore_errors": False}
    if ignore_errors:
        data["ignore_errors"] = ignore_errors
    return client._creoson_post("creo", "set_config", data)


def set_creo_version(client, version):
    """Set the version of Creo you are running.

    This function only needs to be called once per creoson session.
    This function must be called if you are doing certain functions
    in Creo 7 due to deprecated config options.

    Needed for functions:
        familytable_replace
        file_assemble
        file_regenerate
        feature_delete
        feature_resume
        feature_suppress

    Args:
        client (obj):
            creopyson Client.
        version (int):
            Creo version.

    Returns:
        None.

    """
    data = {"version": int(version)}
    return client._creoson_post("creo", "set_creo_version", data)


def set_std_color(client, color_type, red, green, blue):
    """Set one of Creo's standard colors.

    Args:
        client (obj):
            creopyson Client.
        color_type (str):
            Color type.
            Valid values: letter, highlight, drawing, background, half_tone,
            edge_highlight, dimmed, error, warning, sheetmetal, curve,
            presel_highlight, selected, secondary_selected, preview,
            secondary_preview, datum, quilt.
        red (int):
            Red value (0-255).
        green (int):
            Green value (0-255).
        blue (int):
            Blue value (0-255).

    Returns:
        None.

    """
    data = {"color_type": color_type, "red": red, "green": green, "blue": blue}
    return client._creoson_post("creo", "rmdir", data)
    # TODO: convert RGB to a tuple or hexa color?



"""Test server Module."""

import requests
import json
import pytest
import creopyson


def test_server_pwd_ok(monkeypatch):
    """Test server_pwd ok."""
    class Mk_post():
        def __init__(self, *args, **kwargs):
            pass

        def json(self):
            results = {
                "status": {
                    "error": False,
                },
                "data": {
                    "dirname": "C:/CreosonServer-2.3.0-win64"
                }
            }
            return results

    monkeypatch.setattr(requests, 'post', Mk_post)
    c = creopyson.Client()
    result = c.server_pwd()
    assert result == "C:/CreosonServer-2.3.0-win64"


def test_server_pwd_error(monkeypatch):
    """Test creoson return error."""
    class Mk_post():
        def __init__(self, *args, **kwargs):
            pass

        def json(self):
            results = {
                "status": {
                    "error": True,
                    "message": "error message"
                }
            }
            return results

    monkeypatch.setattr(requests, 'post', Mk_post)
    c = creopyson.Client()
    with pytest.raises(Warning) as pytest_wrapped_e:
        c.server_pwd()
    assert pytest_wrapped_e.value.args[0] == "error message"



"""Windchill module.

Connect to Windchill server, use workspaces (create/delete/list)
List files and checkout status.

"""


def authorize(client, user, password):
    """Set user's Windchill login/password.

    Args:
        client (obj): creopyson Client
        user (str): user name
        password (str): password

    Returns:
        None

    """
    data = {"user": user, "password": password}
    return client._creoson_post("windchill", "authorize", data)


def clear_workspace(client, workspace=None, filenames=None):
    """Clear a workspace on the active server.

    Args:
        client (obj):
            creopyson Client
        workspace (str, optionnal):
            Workspace name. Default is current workspace.
        filenames (str|list, optionnal):
            List of files to delete from the workspace.
            Default: All files are deleted.

    Returns:
        None

    """
    active_workspace = client.windchill_get_workspace()
    data = {"workspace": active_workspace}
    if workspace:
        data["workspace"] = workspace
    if filenames:
        data["filenames"] = filenames
    return client._creoson_post("windchill", "clear_workspace", data)


def create_workspace(client, workspace, context_name):
    """Create a workspace on the active server.

    Args:
        client (obj): creopyson Client
        workspace (str): Workspace name
        context_name (str): Context name

    Returns:
        None

    """
    data = {"workspace": workspace, "context": context_name}
    return client._creoson_post("windchill", "create_workspace", data)


def delete_workspace(client, workspace):
    """Delete a workspace on the active server.

    Args:
        client (obj): creopyson Client
        workspace (str): Workspace name

    Returns:
        None

    """
    data = {"workspace": workspace}
    return client._creoson_post("windchill", "delete_workspace", data)


def file_checked_out(client, filename, workspace=None):
    """Check whether a file is checked out in a workspace on the active server.

    Args:
        client (obj):
            creopyson Client
        filename (str):
            File name
        workspace (str, optionnal):
            Workspace name. Default is current workspace.

    Returns:
        Boolean: Whether the file is checked out in the workspace.

    """
    active_workspace = client.windchill_get_workspace()
    data = {"workspace": active_workspace, "filename": filename}
    if workspace:
        data["workspace"] = workspace
    return client._creoson_post("windchill", "file_checked_out", data, "checked_out")


def get_workspace(client):
    """Retrieve the name of the active workspace on the active server.

    Args:
        client (obj): creopyson Client

    Returns:
        str: Active Workspace name.

    """
    return client._creoson_post("windchill", "get_workspace", key_data="workspace")


def list_workspace_files(client, workspace=None, filename=None):
    """Get a list of files in a workspace on the active server.

    Args:
        client (obj):
            creopyson Client
        workspace (str, optionnal):
            Workspace name. Default is current workspace.
        filename (str, optional):
            File name or search. Default is all files.
            ex: `*.asm`, `screw_*.prt`

    Returns:
        list: List of files in the workspace correspnding to the data.

    """
    active_workspace = client.windchill_get_workspace()
    data = {"workspace": active_workspace, "filename": "*"}
    if workspace:
        data["workspace"] = workspace
    if filename:
        data["filename"] = filename
    return client._creoson_post("windchill", "list_workspace_files", data, "filelist")


def list_workspaces(client):
    """Get a list of workspaces the user can access on the active server.

    Args:
        client (obj): creopyson Client

    Returns:
        list: List of workspaces

    """
    return client._creoson_post("windchill", "list_workspaces", key_data="workspaces")


def server_exists(client, server_url):
    """Check whether a server exists.

    Args:
        client (obj): creopyson Client
        server_url (str): server URL or Alias

    Returns:
        Boolean: Whether the server exists

    """
    data = {"server_url": server_url}
    return client._creoson_post("windchill", "server_exists", data, "exists")


def set_server(client, server_url):
    """Select a Windchill server.

    Args:
        client (obj): creopyson Client
        server_url (str): server URL or Alias

    Returns:
        None

    """
    data = {"server_url": server_url}
    return client._creoson_post("windchill", "set_server", data)


def set_workspace(client, workspace):
    """Select a workspace on the active server.

    Args:
        client (obj): creopyson Client
        workspace (str): Workspace name

    Returns:
        None

    """
    data = {"workspace": workspace}
    return client._creoson_post("windchill", "set_workspace", data)


def workspace_exists(client, workspace):
    """Check whether a workspace exists on the active server.

    Args:
        client (obj): creopyson Client
        workspace (str): Workspace name

    Returns:
        Boolean: Whether the workspace exists

    """
    data = {"workspace": workspace}
    return client._creoson_post("windchill", "workspace_exists", data, "exists")



"""Creo testing."""
from .fixtures import mk_creoson_post_dict, mk_getactivefile
import creopyson


def test_bom_get_paths_ok(mk_creoson_post_dict, mk_getactivefile):
    """Test bom_get_paths ok."""
    c = creopyson.Client()
    result = c.bom_get_paths(
        file_="fakefile",
        paths=True,
        skeletons=True,
        top_level=True,
        get_transforms=True,
        exclude_inactive=True
    )
    assert isinstance(result, (dict))
    c = creopyson.Client()
    result = c.bom_get_paths()
    assert isinstance(result, (dict))



"""Feature module."""
STATUS_LIST = [
    "ACTIVE",
    "INACTIVE",
    "FAMILY_TABLE_SUPPRESSED",
    "SIMP_REP_SUPPRESSED",
    "PROGRAM_SUPPRESSED",
    "SUPPRESSED",
    "UNREGENERATED",
]


def delete(client, name=None, file_=None, status=None, type_=None, clip=None):
    """Delete one or more features that match criteria.

    Args:
        client (obj):
            creopyson Client.
        name (str|list:str, optional):
            Dimension name, (wildcards allowed: True);
            if empty then all features are listed.
        `file_` (str, optional):
            Model name (wildcards allowed: True).
            Defaults is current active model.
        status (str, optional):
            Feature status pattern (wildcards allowed: True).
            Defaults: All feature statuses.
            Valid values: ACTIVE, INACTIVE, FAMILY_TABLE_SUPPRESSED,
            SIMP_REP_SUPPRESSED, PROGRAM_SUPPRESSED, SUPPRESSED, UNREGENERATED
        `type_` (str, optional):
            Feature type pattern (wildcards allowed: True).
            Defaults: All feature types.
        clip (boolean, optional):
            Whether to clip-delete ANY features from this feature through
            the end of the structure.
            Defaults is False.

    Raises:
        ValueError: status value is incorrect.

    Returns:
        None

    """
    data = {}
    if file_ is not None:
        data["file"] = file_
    else:
        active_file = client.file_get_active()
        if active_file:
            data["file"] = active_file["file"]
    if name:
        if isinstance(name, (str)):
            data["name"] = name
        elif isinstance(name, (list)):
            data["names"] = name
    if status in STATUS_LIST:
        data["status"] = status
    elif status is not None:
        raise ValueError(f"`{status}` is not a correct status.")
    if type_:
        data["type"] = type_
    if clip:
        data["clip"] = clip
    return client._creoson_post("feature", "delete", data)


def delete_param(client, name=None, file_=None, param=None):
    """Delete a feature parameter.

    Args:
        client (obj):
            creopyson Client.
        name (str, optional):
            Parameter name (wildcards allowed: True).
            Defaults: All parameter names.
        `file_` (str, optional):
            Model name. Defaults is current active model.
        param (str, optional):
            Parameter name (wildcards allowed: True).
            Defaults: All parameter names.

    Returns:
        None

    """
    data = {}
    if file_ is not None:
        data["file"] = file_
    else:
        active_file = client.file_get_active()
        if active_file:
            data["file"] = active_file["file"]
    if name:
        data["name"] = name
    if param:
        data["param"] = param
    return client._creoson_post("feature", "delete_param", data)


def list_(
    client,
    file_=None,
    name=None,
    status=None,
    type_=None,
    paths=None,
    no_datum=None,
    inc_unnamed=None,
    no_comp=None,
):
    """List feature parameters that match criteria.

    Will only list parameters on visible features.

    Args:
        client (obj):
            creopyson Client.
        `file_` (str, optional):
            File name. Defaults is the currently active model.
        name (str, optional):
            Feature name (wildcards allowed: True).
            Defaults: All features are listed.
        status (str, optionnal):
            Feature status pattern.
            Defaults: All feature statuses.
            Valid values: ACTIVE, INACTIVE, FAMILY_TABLE_SUPPRESSED,
            SIMP_REP_SUPPRESSED, PROGRAM_SUPPRESSED, SUPPRESSED, UNREGENERATED.
        `type_` (str, optional):
            Feature type patter (wildcards allowed: True).
            Defaults: All feature types.
        paths (boolean, optionnal):
            Whether feature ID and feature number are returned with the data
            Default: False.
        no_datum (boolean, optional):
            Whether to exclude datum-type features from the list;
            these are COORD_SYS, CURVE, DATUM_AXIS, DATUM_PLANE, DATUM_POINT,
            DATUM_QUILT, and DATUM_SURFACE features.
            Defaults is False.
        inc_unnamed (boolean, optional):
            Whether to include unnamed features in the list.
            Defaults is False.
        no_comp (boolean, optional):
            Whether to include component-type features in the list.
            Defaults is False.

    Raises:
        ValueError: status value is incorrect.

    Returns:
        (list:dict): List of parameter information.
            name (str):
                Parameter nam.
            value (depends on data type):
                Parameter value.
            type (string):
                Data type. Valid values: STRING, DOUBLE, INTEGER, BOOL, NOTE.
            designate (boolean):
                Value is designated.
            encoded (boolean):
                Value is Base64-encoded.
            owner_name (str):
                Owner Name.
            owner_id (int):
                Owner ID.
            owner_type (str):
                Owner type.

    """
    data = {}
    if file_ is not None:
        data["file"] = file_
    else:
        active_file = client.file_get_active()
        if active_file:
            data["file"] = active_file["file"]
    if name:
        data["name"] = name
    if status in STATUS_LIST:
        data["status"] = status
    elif status is not None:
        raise ValueError(f"`{status}` is not a correct status.")
    if type_:
        data["type"] = type_
    if paths:
        data["paths"] = paths
    if no_datum:
        data["no_datum"] = no_datum
    if inc_unnamed:
        data["inc_unnamed"] = inc_unnamed
    if no_comp:
        data["no_comp"] = no_comp
    return client._creoson_post("feature", "list", data, "featlist")


def list_params(
    client,
    file_=None,
    name=None,
    type_=None,
    no_datum=None,
    inc_unnamed=None,
    no_comp=None,
    param=None,
    value=None,
    encoded=None,
):
    """List feature parameters that match criteria.

    Args:
        client (obj):
            creopyson Client.
        `file_` (str, optional):
            File name. Defaults is the currently active model.
        name (str|int, optional):
            str: Feature name (wildcards allowed: True).
            int: Feature ID.
            Defaults: All features are listed.
        `type_` (str, optional):
            Feature type patter (wildcards allowed: True).
            Defaults: All feature types.
        no_datum (boolean, optional):
            Whether to exclude datum-type features from the list;
            these are COORD_SYS, CURVE, DATUM_AXIS, DATUM_PLANE, DATUM_POINT,
            DATUM_QUILT, and DATUM_SURFACE features.
            Defaults is False.
        inc_unnamed (boolean, optional):
            Whether to include unnamed features in the list.
            Defaults is False.
        no_comp (boolean, optional):
            Whether to include component-type features in the list.
            Defaults is False.
        param (str|list:str, optional):
            Parameter name; (wildcards allowed: True)
            if empty all parameters are listed.
        value (str, optional):
            Parameter value filter (wildcards allowed: True).
            Defaults is no filter.
        encoded (boolean, optional):
            Whether to return the values Base64-encoded.
            Defaults is False.

    Returns:
        (list:dict): List of parameter information.
            name (str):
                Parameter nam.
            value (depends on data type):
                Parameter value.
            type (string):
                Data type. Valid values: STRING, DOUBLE, INTEGER, BOOL, NOTE.
            designate (boolean):
                Value is designated.
            encoded (boolean):
                Value is Base64-encoded.
            owner_name (str):
                Owner Name.
            owner_id (int):
                Owner ID.
            owner_type (str):
                Owner type.
            description (str):
                List of parameter information.

    """
    data = {}
    if file_ is not None:
        data["file"] = file_
    else:
        active_file = client.file_get_active()
        if active_file:
            data["file"] = active_file["file"]
    if name:
        if isinstance(param, (str)):
            data["name"] = name
        elif isinstance(param, (list)):
            data["feat_id"] = name
    if type_:
        data["type"] = type_
    if no_datum:
        data["no_datum"] = no_datum
    if inc_unnamed:
        data["inc_unnamed"] = inc_unnamed
    if no_comp:
        data["no_comp"] = no_comp
    if param:
        if isinstance(param, (str)):
            data["param"] = param
        elif isinstance(param, (list)):
            data["params"] = param
    if value:
        data["value"] = value
    if encoded:
        data["encoded"] = encoded
    return client._creoson_post("feature", "list_params", data, "paramlist")


def list_group_features(client, group_name, type_=None, file_=None):
    """List features in a Creo Group.

    Args:
        client (obj):
            creopyson Client.
        group_name (str):
            Group name.
        `type_` (str, optional):
            Feature type patter (wildcards allowed: True).
            Defaults: All feature types.
        `file_` (str, optional):
            File name. Defaults is the currently active model.

    Returns:
        (list:dict): List of feature information

    """
    data = {
        "group_name": group_name,
    }
    if file_ is not None:
        data["file"] = file_
    else:
        active_file = client.file_get_active()
        if active_file:
            data["file"] = active_file["file"]
    if type_:
        data["type"] = type_
    return client._creoson_post("feature", "list_group_features", data, "featlist")


def list_pattern_features(client, patter_name, type_=None, file_=None):
    """List features in a Creo Pattern.

    Args:
        client (obj):
            creopyson Client.
        patter_name (str):
            Pattern name.
        `type_` (str, optional):
            Feature type patter (wildcards allowed: True).
            Defaults: All feature types.
        `file_` (str, optional):
            File name. Defaults is the currently active model.

    Returns:
        (list:dict): List of feature information

    """
    data = {
        "patter_name": patter_name,
    }
    if file_ is not None:
        data["file"] = file_
    else:
        active_file = client.file_get_active()
        if active_file:
            data["file"] = active_file["file"]
    if type_:
        data["type"] = type_
    return client._creoson_post("feature", "list_group_features", data, "featlist")


def list_selected(client):
    """List the currently selected features in Creo

    Returns:
        (list): List of feature informations.
            [
                {
                    'file' : model name (str)

                    'name' : feature name (str)

                    'status' : feature status (str)

                    'type' : feature type (str)

                    'feat_id' : feature ID (int)

                    'feat_number' : feature number (int)

                    'path' : feature's component path (list of ints)

                },

            ]

    """
    data = None
    return client._creoson_post("feature", "list_selected", data, "featlist")


def param_exists(client, file_=None, name=None, param=None):
    """Check whether parameter(s) exists on a feature.

    Args:
        client (obj):
            creopyson Client.
        `file_` (str, optional):
            File name. Defaults is the currently active model.
        name (str, optional):
            Parameter name (wildcards allowed: True).
            Defaults: All parameter names.
        param (str|list:str, optional):
            Parameter name; (wildcards allowed: True)
            if empty all parameters are listed.

    Returns:
        (boolean): Whether the parameter exists on the model

    """
    data = {}
    if file_ is not None:
        data["file"] = file_
    else:
        active_file = client.file_get_active()
        if active_file:
            data["file"] = active_file["file"]
    if name:
        data["name"] = name
    if param:
        if isinstance(param, (str)):
            data["param"] = param
        elif isinstance(param, (list)):
            data["params"] = param
    return client._creoson_post("feature", "param_exists", data, "exists")


def rename(client, name, new_name, file_=None):
    """Rename a feature.

    Args:
        client (obj):
            creopyson Client.
        name (str|int, optional):
            Feature name (str) or Feature ID (int).
        new_name (str):
            New name for the feature.
        `file_` (str, optional):
            File name.
            Defaults is the currently active model.

    Returns:
        None

    """
    data = {"new_name": new_name}
    if file_ is not None:
        data["file"] = file_
    else:
        active_file = client.file_get_active()
        if active_file:
            data["file"] = active_file["file"]
    if isinstance(name, (str)):
        data["name"] = name
    elif isinstance(name, (int)):
        data["feat_id"] = name
    else:
        raise TypeError("name must be str or int")
    return client._creoson_post("feature", "rename", data)


def resume(client, file_=None, name=None, status=None, type_=None, with_children=None):
    """Resume one or more features that match criteria.

    Will only resume visible features.

    Args:
        client (obj):
            creopyson Client.
        `file_` (str, optional):
            File name (wildcards allowed: True).
            Defaults is the currently active model.
        name (int|str|list:str, optional):
            Feature name or Feature ID, (wildcards allowed: True);
            if empty then all features are resumed.
            int => Feat_ID
            str => name
            list:str => names
        status (str, optional):
            Feature status pattern. Defaults: All feature statuses.
            Valid values: ACTIVE, INACTIVE, FAMILY_TABLE_SUPPRESSED,
            SIMP_REP_SUPPRESSED, PROGRAM_SUPPRESSED, SUPPRESSED, UNREGENERATED
        `type_` (str, optional):
            Feature type pattern (wildcards allowed: True).
            Defaults: All feature types.
        with_children (boolean, optional):
            Whether to resume any child features of the resumed feature.
            Defaults is False.

    Raises:
        ValueError: status value is incorrect.

    Returns:
        None

    """
    data = {
        "with_children": False,
    }
    if file_ is not None:
        data["file"] = file_
    else:
        active_file = client.file_get_active()
        if active_file:
            data["file"] = active_file["file"]
    if name:
        if isinstance(name, (int)):
            data["feat_id"] = name
        elif isinstance(name, (str)):
            data["name"] = name
        elif isinstance(name, (list)):
            data["names"] = name
    if status in STATUS_LIST:
        data["status"] = status
    elif status is not None:
        raise ValueError(f"`{status}` is not a correct status.")
    if type_:
        data["type"] = type_
    if with_children:
        data["with_children"] = with_children
    return client._creoson_post("feature", "resume", data)


def set_param(
    client,
    param,
    file_=None,
    name=None,
    type_=None,
    value=None,
    encoded=None,
    designate=None,
    description=None,
    no_create=None,
):
    """Set the value of a feature parameter.

    Will only set parameters on visible features.

    Args:
        client (obj):
            creopyson Client.
        param (str):
            Parameter name.
        `file_` (str, optional):
            File name (wildcards allowed: True).
            Defaults is the currently active model.
        name (str, optional):
            Feature name. Defaults: All features are updated.
        `type_` (str, optional):
            Parameter data type. Defaults is True.
            Valid values: STRING, DOUBLE, INTEGER, BOOL, NOTE.
        value (depends on data type, optional):
            Parameter value. Defaults: Clears the parameter value if missing.
        encoded (boolean, optional):
            Value is Base64-encoded. Defaults is False.
        designate (boolean, optional):
            Set parameter to be designated/not designated, blank=do not set.
            Defaults is `blank`.
        description (str, optionnal):
            Parameter description. If missing, leaves the currect description in place.
        no_create (boolean, optional):
            If parameter does not already exist, do not create it.
            Defaults is False.

    Returns:
        None

    """
    data = {}
    if file_ is not None:
        data["file"] = file_
    else:
        active_file = client.file_get_active()
        if active_file:
            data["file"] = active_file["file"]
    if name:
        data["name"] = name
    if param:
        data["param"] = param
    if type_:
        data["type"] = type_
    if value:
        data["value"] = value
    if encoded:
        data["encoded"] = encoded
    if designate:
        data["designate"] = designate
    if description:
        data["description"] = description
    if no_create:
        data["no_create"] = no_create
    return client._creoson_post("feature", "set_param", data)


def suppress(
    client,
    file_=None,
    name=None,
    status=None,
    type_=None,
    clip=None,
    with_children=None,
):
    """Suppress one or more features that match criteria.

    Will only suppress visible features.

    Args:
        client (obj):
            creopyson Client.
        `file_` (str, optional):
            File name (wildcards allowed: True).
            Defaults is the currently active model.
        name (int|str|list:str, optional):
            Feature name or Feature ID, (wildcards allowed: True);
            if empty then all features are suppressed.
            int => Feat_ID
            str => name
            list:str => names
        status (str, optional):
            Feature status pattern. Defaults: All feature statuses.
            Valid values: ACTIVE, INACTIVE, FAMILY_TABLE_SUPPRESSED,
            SIMP_REP_SUPPRESSED, PROGRAM_SUPPRESSED, SUPPRESSED, UNREGENERATED
        `type_` (str, optional):
            Feature type pattern (wildcards allowed: True).
            Defaults: All feature types.
        clip (boolean, optional):
            Whether to clip-suppress ANY features from this feature through
            the end of the structure. Defaults is True.
        with_children (boolean, optional):
            Whether to suppress  any child features of the suppressed feature.
            Defaults is True.

    Raises:
        ValueError: status value is incorrect.

    Returns:
        None

    """
    data = {"clip": True, "with_children": True}
    if file_ is not None:
        data["file"] = file_
    else:
        active_file = client.file_get_active()
        if active_file:
            data["file"] = active_file["file"]
    if name:
        if isinstance(name, (int)):
            data["feat_id"] = name
        elif isinstance(name, (str)):
            data["name"] = name
        elif isinstance(name, (list)):
            data["names"] = name
    if status in STATUS_LIST:
        data["status"] = status
    elif status is not None:
        raise ValueError(f"`{status}` is not a correct status.")
    if type_:
        data["type"] = type_
    if clip is False:
        data["clip"] = False
    if with_children is False:
        data["with_children"] = False
    return client._creoson_post("feature", "suppress", data)


def user_select_csys(client, file_=None, max_=None):
    """Prompt the user to select one or more coordinate systems.

    and return their selections.

    Args:
        client (obj):
            creopyson Client.
        `file_` (str, optional):
            File name.
            Defaults is the currently active model.
        `max_` (int, optional):
            The maximum number of dimensions that the user can select.
            Defaults is `1`.

    Returns:
        (list:dict):
            List of feature information.
                name (str):
                    Feature name.
                type (string):
                    Feature type.
                status (str):
                    Feature status.
                    Valid values: ACTIVE, INACTIVE, FAMILY_TABLE_SUPPRESSED,
                    SIMP_REP_SUPPRESSED, PROGRAM_SUPPRESSED, SUPPRESSED,
                    UNREGENERATED.
                feat_id (int):
                    Feature ID.
                file (str):
                    File name containing the feature.
                path (list:int):
                    Component Path to feature (optionnal)

    """
    data = {
        "max": 1,
    }
    if file_ is not None:
        data["file"] = file_
    else:
        active_file = client.file_get_active()
        if active_file:
            data["file"] = active_file["file"]
    if max_:
        data["max"] = max_
    return client._creoson_post("feature", "user_select_csys", data)



"""Note module."""


def copy(client, name, to_name=None, file_=None, to_file=None):
    """Copy note to another in the same model or another model.

    Args:
        client (obj):
            creopyson Client.
        name (str):
            Note name to copy (wildcards allowed: True).
        to_name (str):
            Destination note.
            Defaults is the source note name
        `file_` (str, optional):
            Model name (wildcards allowed: True).
            Defaults is current active model.
        to_file (str, optional):
            Destination model.
            Defaults is the source model.

    Returns:
        None

    """
    data = {"name": name}
    if to_name:
        data["to_name"] = to_name
    if file_ is not None:
        data["file"] = file_
    else:
        active_file = client.file_get_active()
        if active_file:
            data["file"] = active_file["file"]
    if to_file:
        data["to_file"] = to_file
    return client._creoson_post("note", "copy", data)


def delete(client, name, file_=None):
    """Delete a model or drawing note.

    Args:
        client (obj):
            creopyson Client.
        name (str):
            Note name (wildcards allowed: True).
        `file_` (str, optional):
            Model name (wildcards allowed: True).
            Defaults is current active model.

    Returns:
        None

    """
    data = {"name": name}
    if file_ is not None:
        data["file"] = file_
    else:
        active_file = client.file_get_active()
        if active_file:
            data["file"] = active_file["file"]
    return client._creoson_post("note", "delete", data)


def exists(client, file_=None, name=None):
    """Check whether note(s) exists on a model.

    Args:
        client (obj):
            creopyson Client.
        `file_` (str, optional):
            Model name. Defaults is current active model.
        name (str|list:str, optional):
            Note name; List of note names.
            if empty it checks for any note's existence.

    Returns:
        (boolean): Whether the note exists on the model.

    """
    data = {}
    if file_ is not None:
        data["file"] = file_
    else:
        active_file = client.file_get_active()
        if active_file:
            data["file"] = active_file["file"]
    if name:
        if isinstance(name, (str)):
            data["name"] = name
        elif isinstance(name, (list)):
            data["names"] = name
    return client._creoson_post("note", "exists", data, "exists")


def get(client, name, file_=None):
    """Get the text of a model or drawing note.

    Values will automatically be returned Base64-encoded if they are strings
    which contain Creo Symbols or other non-ASCII data.

    Args:
        client (obj):
            creopyson Client.
        name (str):
            Note name.
        `file_` (str, optional):
            Model name.
            Defaults is current active model or drawing.

    Returns:
        (dict):
            file (str):
                File name.
            name (str):
                Note name.
            encoded (boolean):
                Value is Base64-encoded.
            url (str):
                "Note URL, if there is one.
            location (obj:JLPoint):
                Note location in Drawing Units (drawing notes only)

    """
    data = {"name": name}
    if file_ is not None:
        data["file"] = file_
    else:
        active_file = client.file_get_active()
        if active_file:
            data["file"] = active_file["file"]
    return client._creoson_post("note", "get", data)


def list_(client, file_=None, name=None, value=None, get_expanded=None, select=False):
    """Get a list of notes from one or more models.

    Values will automatically be returned Base64-encoded if they are strings
    which contain Creo Symbols or other non-ASCII data.

    Args:
        client (obj):
            creopyson Client.
        `file_` (str, optional):
            Model name (wildcards allows: True).
            Defaults is current active model.
        name (str|list:str, optional):
            Note name; List of note names.
            if empty all notes are listed.
        value (str, optional):
            Parameter value filter (wildcards allows: True).
            Defaults is `no filter`.
        get_expanded (boolean, optional):
            Whether to return text with parameter values replaced.
            Defaults is False.
        select (boolean, optional):
            If true, the notes that are found will be selected in Creo.
            Defaults is False.

    Returns:
        (list:dict):
            name (str): Note name.
            value (str): Note text with parameters not expanded.
            value_expanded (str): Note text with parameters expanded.
            encoded (boolean): Value is Base64-encoded.
            location (jlpoint): 3D coordinate dict.

    """
    data = {}
    if file_ is not None:
        data["file"] = file_
    else:
        active_file = client.file_get_active()
        if active_file:
            data["file"] = active_file["file"]
    if name:
        if isinstance(name, (str)):
            data["name"] = name
        elif isinstance(name, (list)):
            data["names"] = name
    if get_expanded:
        data["get_expanded"] = True
    if value:
        data["value"] = value
    if select:
        data["select"] = True
    return client._creoson_post("note", "list", data, "itemlist")


def set_(client, name, file_=None, encoded=None, value=None):
    r"""Set the text of a model or drawing note.

    The location parameter can used to position a new note, or to change the
    position of an existing note. If the text contains Creo Symbols or other
    non-ASCII text, you must Base64-encode the value and set encoded to true
    You may be able to avoid Base64-encoding symbols by using Unicode for the
    binary characters, for example including \\u0001#\\u0002 in the value to
    insert a plus/minus symbol. Embed newlines in the value for line breaks.

    Args:
        client (obj):
            creopyson Client.
        name (str):
            Note name (wildcards allowed: True).
        `file_` (str, optional):
            Model name. Defaults is current active model or drawing.
        encoded (boolean, optional):
            Whether the value is Base64-encoded. Defaults is False.
        value (str, optional):
            Note text with parameters not expanded.
            Defaults to None: clears the note if missing.

    Returns:
        None

    """
    data = {"name": name}
    if file_ is not None:
        data["file"] = file_
    else:
        active_file = client.file_get_active()
        if active_file:
            data["file"] = active_file["file"]
    if encoded:
        data["encoded"] = encoded
    if value:
        data["value"] = value
    return client._creoson_post("note", "set", data)



"""Objects Tests."""
import pytest
from creopyson.objects import jlpoint


def test_objects_jlpoint():
    """Test jlpoint."""
    result = jlpoint(1, 2, 3)
    assert result == {'x': 1.0, 'y': 2.0, 'z': 3.0}
    # Test fail
    with pytest.raises(Warning):
        jlpoint(1, "2", "trois")



"""Creopyson exceptions."""


class Error(Exception):
    """Base class for other exceptions."""

    pass


class MissingKey(Error):
    """Raised when the input value is too small."""

    pass


class ErrorJsonDecode(Error):
    """Raised when creoson result cannot be decoded."""

    pass



"""Bom module."""


def get_paths(
    client,
    file_=None,
    paths=None,
    skeletons=None,
    top_level=None,
    get_transforms=None,
    exclude_inactive=None,
):
    """Get a hierarchy of components within an assembly.

    Even if you do not set exclude_inactive to true, the function will still
    exclude any components with a status of INACTIVE or UNREGENERATED.

    Args:
        client (obj):
            creopyson Client.
        `file_` (string, optional):
            file name, if not set, active model is used.
        paths (boolean, optional):
            Whether to return component paths for each component
            (default" : False)
        skeletons (boolean, optional):
            Whether to include skeleton components
            (default" : False)
        top_level (boolean, optional):
            Whether to return only the top-level components
            in the assembly. (default" : False)
        get_transforms (boolean, optional):
            Whether to return the 3D transform matrix
            for each component. (default" : False)
        exclude_inactive (boolean, optional):
            Whether to exclude components which do not
            have an ACTIVE status. (default" : False)

    Returns:
        Dict:
            file (string):
                Assembly file name
            generic (string):
                Generic name for the assembly
            children (object:BomChild):
                The hierarchy of component data,
                starting with the top-level assembly.
            has_simprep (boolean):
                Whether the assembly has a Simplified Rep.

            in `children` there is the `seq_path` which indicates
            the children level, ex: `root.3.2`.

    """
    data = {}
    if file_ is not None:
        data["file"] = file_
    else:
        active_file = client.file_get_active()
        if active_file:
            data["file"] = active_file["file"]
    if paths:
        data["paths"] = paths
    if skeletons:
        data["skeletons"] = skeletons
    if top_level:
        data["top_level"] = top_level
    if get_transforms:
        data["get_transforms"] = get_transforms
    if exclude_inactive:
        data["exclude_inactive"] = exclude_inactive
    return client._creoson_post("bom", "get_paths", data)



"""Layer module."""


def delete(client, name=None, file_=None):
    """Delete one or more layers.

    Args:
        client (obj):
            creopyson Client.
        name (str, optional):
            Layer name (wildcards allowed: True).
            Defaults: All layers will be deleted.
        `file_` (str, optional):
            File name (wildcards alloawed: True).
            Defaults is current active model.

    Returns:
        None

    """
    data = {"name": name}
    if file_ is not None:
        data["file"] = file_
    else:
        active_file = client.file_get_active()
        if active_file:
            data["file"] = active_file["file"]
    if name:
        data["name"] = name
    return client._creoson_post("layer", "delete", data)


def exists(client, name=None, file_=None):
    """Check whether layer(s) exists on a model.

    Args:
        client (obj):
            creopyson Client.
        name (str, optional):
            Layer name (wildcards allowed: True).
            Defaults: All layers are listed.
        `file_` (str, optional):
            File name.
            Defaults is current active model.

    Returns:
        (boolean): Whether the layer exists on the model.

    """
    data = {}
    if file_ is not None:
        data["file"] = file_
    else:
        active_file = client.file_get_active()
        if active_file:
            data["file"] = active_file["file"]
    if name:
        data["name"] = name
    return client._creoson_post("layer", "exists", data, "exists")


def list_(client, name=None, file_=None):
    """List layers that match criteria.

    Args:
        client (obj):
            creopyson Client.
        name (str, optional):
            Layer name (wildcards allowed: True).
            Defaults: All layers are listed.
        `file_` (str, optional):
            File name.
            Defaults is current active model.

    Returns:
        (list:dict):
            name (str):
                Layer name.
            status (str): Layer status.
                Valid values: BLANK, DISPLAY, HIDDEN, NORMAL.
            ID (int):
                Layer ID.

    """
    data = {}
    if file_ is not None:
        data["file"] = file_
    else:
        active_file = client.file_get_active()
        if active_file:
            data["file"] = active_file["file"]
    if name:
        data["name"] = name
    return client._creoson_post("layer", "list", data, "layers")


def show(client, name=None, file_=None, show_=None):
    """how/Hide one or more layers.

    Args:
        client (obj):
            creopyson Client.
        name (str, optional):
            Layer name (wildcards allowed: True).
            Defaults: All layers are listed.
        `file_` (str, optional):
            File name (wildcards allowed: True).
            Defaults is current active model.
        `show_` (boolean, optional):
            Whether to show or hide the layers.
            Defaults is True (show).

    Returns:
        None

    """
    data = {}
    if file_ is not None:
        data["file"] = file_
    else:
        active_file = client.file_get_active()
        if active_file:
            data["file"] = active_file["file"]
    if name:
        data["name"] = name
    if show_:
        data["show"] = show_
    return client._creoson_post("layer", "show", data)



"""Familytable module."""


def add_inst(client, instance, file_=None):
    """Add a new instance to a family table.

    Creates a family table if one does not exist.

    Args:
        client (obj):
            creopyson Client.
        instance (str):
            New instance name.
        `file_` (str, optional):
            File name. Defaults is currently active model.

    Returns:
        None

    """
    data = {"instance": instance}
    if file_ is not None:
        data["file"] = file_
    else:
        active_file = client.file_get_active()
        if active_file:
            data["file"] = active_file["file"]
    return client._creoson_post("familytable", "add_inst", data)


def create_inst(client, instance, file_=None):
    """Create a new model from a family table row.

    Args:
        client (obj):
            creopyson Client.
        instance (str):
            Instance name.
        `file_` (str, optional):
            File name. Defaults is currently active model.

    Returns:
        (str): New model name.

    """
    data = {"instance": instance}
    if file_ is not None:
        data["file"] = file_
    else:
        active_file = client.file_get_active()
        if active_file:
            data["file"] = active_file["file"]
    return client._creoson_post("familytable", "create_inst", data, "name")


def delete_inst(client, instance, file_=None):
    """Delete an instance from a family table.

    Args:
        client (obj):
            creopyson Client.
        instance (str):
            Instance name.
        `file_` (str, optional):
            File name. Defaults is currently active model.

    Returns:
        None

    """
    data = {"instance": instance}
    if file_ is not None:
        data["file"] = file_
    else:
        active_file = client.file_get_active()
        if active_file:
            data["file"] = active_file["file"]
    return client._creoson_post("familytable", "delete_inst", data)


def delete(client, file_=None):
    """Delete an entire family table.

    Args:
        client (obj):
            creopyson Client.
        `file_` (str, optional):
            File name (wildcards allowed: True).
            Defaults is currently active model.

    Returns:
        None

    """
    data = {}
    if file_ is not None:
        data["file"] = file_
    else:
        active_file = client.file_get_active()
        if active_file:
            data["file"] = active_file["file"]
    return client._creoson_post("familytable", "delete", data)


def exists(client, instance, file_=None):
    """Check whether an instance exists in a family table.

    Args:
        client (obj):
            creopyson Client.
        instance (str):
            Instance name.
        `file_` (str, optional):
            File name. Defaults is currently active model.

    Returns:
        (boolean):
            Whether the instance exists in the model's family table;
            returns false if there is no family table in the model.

    """
    data = {"instance": instance}
    if file_ is not None:
        data["file"] = file_
    else:
        active_file = client.file_get_active()
        if active_file:
            data["file"] = active_file["file"]
    return client._creoson_post("familytable", "exists", data, "exists")


def get_cell(client, instance, colid, file_=None):
    """Get one cell of a family table.

    Args:
        client (obj):
            creopyson Client.
        instance (str):
            Instance name.
        colid (str):
            Colimn ID.
        `file_` (str, optional):
            File name. Defaults is currently active model.

    Returns:
        (dict):
            instance (str):
                Family Table instance name.
            colid (str):
                Column ID.
            value (depends on datatype):
                Cell value.
            datatype (str):
                Data type.
            coltype (str):
                Column Type; a string corresponding to the Creo column type.

    """
    data = {"instance": instance, "colid": colid}
    if file_ is not None:
        data["file"] = file_
    else:
        active_file = client.file_get_active()
        if active_file:
            data["file"] = active_file["file"]
    return client._creoson_post("familytable", "get_cell", data)


def get_header(client, file_=None):
    """Get the header of a family table.

    Args:
        client (obj):
            creopyson Client.
        `file_` (str, optional):
            File name.
            Defaults is currently active model.

    Returns:
        (list:dict):
            colid (str):
                Column ID.
            value (depends on date type):
                Cell value.
            datatype (str):
                Data type. Valid values: STRING, DOUBLE, INTEGER, BOOL, NOTE.
            coltype (str):
                Column Type; a string corresponding to the Creo column type.

    """
    data = {}
    if file_ is not None:
        data["file"] = file_
    else:
        active_file = client.file_get_active()
        if active_file:
            data["file"] = active_file["file"]
    return client._creoson_post("familytable", "get_header", data, "columns")


def get_parents(client, file_=None):
    """Get the parent instances of a model in a nested family table.

    Args:
        client (obj):
            creopyson Client.
        `file_` (str, optional):
            File name.
            Defaults is currently active model.

    Returns:
        (list:str):
            List of parent instance names, starting with
            the immediate parent and working back.

    """
    data = {}
    if file_ is not None:
        data["file"] = file_
    else:
        active_file = client.file_get_active()
        if active_file:
            data["file"] = active_file["file"]
    return client._creoson_post("familytable", "get_parents", data, "parents")


def get_row(client, instance, file_=None):
    """Get one row of a family table.

    Args:
        client (obj):
            creopyson Client.
        instance (str):
            Instance name.
        `file_` (str, optional):
            File name. Defaults is currently active model.

    Returns:
        (dict):
            colid (str):
                Column ID.
            value (depends on datatype):
                Cell value.
            datatype (str):
                Data type.
            coltype (str):
                Column Type; a string corresponding to the Creo column type.

    """
    data = {"instance": instance}
    if file_ is not None:
        data["file"] = file_
    else:
        active_file = client.file_get_active()
        if active_file:
            data["file"] = active_file["file"]
    return client._creoson_post("familytable", "get_row", data, "columns")


def list_(client, file_=None, instance=None):
    """List the instance names in a family table.

    Args:
        client (obj):
            creopyson Client.
        instance (str, optional):
            Instance name filter (wildcards allowed: True).
            Defaults is all instances.
        `file_` (str, optional):
            File name. Defaults is currently active model.

    Returns:
        (list:str): List of matching instance names

    """
    data = {}
    if file_ is not None:
        data["file"] = file_
    else:
        active_file = client.file_get_active()
        if active_file:
            data["file"] = active_file["file"]
    if instance:
        data["instance"] = instance
    return client._creoson_post("familytable", "list", data, "instances")


def list_tree(client, file_=None, erase=None):
    """Get a hierarchical structure of a nested family table.

    Args:
        client (obj):
            creopyson Client.
        `file_` (str, optional):
            File name. Defaults is currently active model.
        erase (boolean, optional):
            Erase model and non-displayed models afterwards.
            Defaults is `False`.

    Returns:
        (list:str):
            List of child instances
                total (int):
                    Count of all child instances including their decendants.
                children (list:dict):
                    name (str):
                        Name of the family table instance.
                    total (int):
                        Count of all child instances including their
                        decendants.
                    children (list:dict):
                        List of child instances.

    """
    data = {}
    if file_ is not None:
        data["file"] = file_
    else:
        active_file = client.file_get_active()
        if active_file:
            data["file"] = active_file["file"]
    if erase:
        data["erase"] = erase
    return client._creoson_post("familytable", "list_tree", data, "children")


def replace(client, cur_model, new_inst, file_=None, cur_inst=None, path=None):
    """Replace a model in an assembly with another inst in the same family table.

    You must specify either cur_inst or path.

    Args:
        client (obj):
            creopyson Client.
        cur_model (str):
            Generic model containing the instances.
        new_inst (str):
            New instance name.
        `file_` (str, optional):
            File name (usually an assembly).
            Defaults is currently active model.
        cur_inst (str):
            Instance name to replace. Defaults to None.
        path (list:int, optional):
            Path to component to replace. Defaults to None.

    Returns:
        None

    """
    data = {
        "cur_model": cur_model,
        "new_inst": new_inst,
    }
    if file_ is not None:
        data["file"] = file_
    else:
        active_file = client.file_get_active()
        if active_file:
            data["file"] = active_file["file"]
    if cur_inst:
        data["cur_inst"] = cur_inst
    if path:
        data["path"] = path
    return client._creoson_post("familytable", "replace", data)
    # TODO: path/cur_inst


def set_cell(client, instance, colid, value, file_=None):
    """Set the value of one cell of a family table.

    Args:
        client (obj):
            creopyson Client.
        instance (str):
            Family Table instance name.
        colid (str):
            Column ID.
        value (depends on data type):
            Cell value.
        `file_` (str, optional):
            File name (usually an assembly).
            Defaults is currently active model.

    Returns:
        None

    """
    data = {
        "instance": instance,
        "colid": colid,
        "value": value,
    }
    if file_ is not None:
        data["file"] = file_
    else:
        active_file = client.file_get_active()
        if active_file:
            data["file"] = active_file["file"]
    return client._creoson_post("familytable", "set_cell", data)



"""Geometry module."""


def bound_box(client, file_=None):
    """Get the bounding box for a model.

    Args:
        client (obj):
            creopyson Client.
        `file_` (str, optional):
            File name.
            Defaults is current active model.

    Returns:
        (dict):
            xmin (float): Minimum X-coordinate of model.
            xmax (float): Maximum X-coordinate of model.
            ymin (float): Minimum Y-coordinate of model.
            ymax (float): Maximum Y-coordinate of model
            zmin (float): Minimum Z-coordinate of model.
            zmax (float): Maximum Z-coordinate of model

    """
    data = {}
    if file_ is not None:
        data["file"] = file_
    else:
        active_file = client.file_get_active()
        if active_file:
            data["file"] = active_file["file"]
    return client._creoson_post("geometry", "bound_box", data)


def get_edges(client, surface_ids, file_=None):
    """Get the list of edges for a model for given surfaces.

    Args:
        client (obj):
            creopyson Client.
        surface_ids (list:int):
            List of surface IDs.
        `file_` (str, optional):
            File name.
            Defaults is current active model

    Returns:
        (list:dict):
            surface_id (int):
                Surface ID.
            traversal (string):
                Traversal type. Valid values: internal, external.
            edglist (list:dict):
                Information about an edge.
                    edgeid (integer):
                        Edge ID.
                    length (float):
                        Edge length.
                    start (list:dict):
                        A 3D coordinate.
                            x (float):
                                X-coordinate.
                            y (float):
                                Y-coordinate.
                            z (float):
                                Z-coordinate.
                    end (list:dict):
                        A 3D coordinate.
                            x (float):
                                X-coordinate.
                            y (float):
                                Y-coordinate.
                            z (float):
                                Z-coordinate.

    """
    data = {"surface_ids": surface_ids}
    if file_ is not None:
        data["file"] = file_
    else:
        active_file = client.file_get_active()
        if active_file:
            data["file"] = active_file["file"]
    return client._creoson_post("geometry", "get_edges", data, "contourlist")


def get_surfaces(client, file_=None):
    """Get the list of surfaces for a model.

    Args:
        client (obj):
            creopyson Client.
        `file_` (str, optional):
            File name.
            Defaults is current active model

    Returns:
        (list:dict):
            surface_id (int):
                Surface ID.
            area (float):
                Surface area
            min_extent (list:dict):
                A 3D coordinate.
                    x (float):
                        X-coordinate.
                    y (float):
                        Y-coordinate.
                    z (float):
                        Z-coordinate.
            max_extent (list:dict):
                A 3D coordinate.
                    x (float):
                        X-coordinate.
                    y (float):
                        Y-coordinate.
                    z (float):
                        Z-coordinate.

    """
    data = {}
    if file_ is not None:
        data["file"] = file_
    else:
        active_file = client.file_get_active()
        if active_file:
            data["file"] = active_file["file"]
    return client._creoson_post("geometry", "get_surfaces", data, "surflist")



"""Core testing."""

from creopyson.core import creoson_post
import creopyson
import requests
import json


def test_core_creoson_post_return_error_msg(monkeypatch):
    """Test when creoson return an error message."""
    class Mk_post():
        def __init__(self, *args, **kwargs):
            pass

        @property
        def content(self):
            results = {
                "status": {
                    "error": True,
                    "message": "Error message"
                },
                "data": None
            }
            return json.dumps(results).encode()

    monkeypatch.setattr(requests, 'post', Mk_post)

    c = creopyson.Client()

    status, data = creoson_post(c, None)

    assert status
    assert data == "Error message"


def test_core_creoson_post_return_data(monkeypatch):
    """Test when creoson return data."""
    class Mk_post():
        def __init__(self, *args, **kwargs):
            pass

        @property
        def content(self):
            results = {
                "status": {
                    "error": False,
                    "message": "Error message"
                },
                "data": {
                    "random_data": 12345,
                }
            }
            return json.dumps(results).encode()

    monkeypatch.setattr(requests, 'post', Mk_post)

    c = creopyson.Client()

    status, data = creoson_post(c, None)

    assert not status
    assert data
    assert isinstance(data, dict)


def test_core_creoson_post_return_None(monkeypatch):
    """Test when creoson return nothing."""
    class Mk_post():
        def __init__(self, *args, **kwargs):
            pass

        @property
        def content(self):
            results = {
                "status": {
                    "error": False,
                    "message": "Error message"
                },
            }
            return json.dumps(results).encode()

    monkeypatch.setattr(requests, 'post', Mk_post)

    c = creopyson.Client()

    status, data = creoson_post(c, None)

    assert not status
    assert data is None



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

requirements = ["requests"]

setup_requirements = [
    "pytest-runner",
]

test_requirements = [
    "pytest",
]

setup(
    author="Benjamin C.",
    author_email="zepman@gmail.com",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    description="Python library for Creoson",
    install_requires=requirements,
    license="MIT license",
    long_description=readme + "\n\n" + history,
    include_package_data=True,
    keywords="creopyson",
    name="creopyson",
    packages=find_packages(include=["creopyson"]),
    python_requires=">=3.7",
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/Zepmanbc/creopyson",
    version="0.7.2",
    zip_safe=False,
)



"""View module."""


def activate(client, name, file_=None):
    """Activate a model view.

    Args:
        client (obj):
            creopyson Client.
        name (str):
            View name.
        `file_` (str, optional):
            Model name. Defaults is current active model.

    Returns:
        None

    """
    data = {"name": name}
    if file_ is not None:
        data["file"] = file_
    else:
        active_file = client.file_get_active()
        if active_file:
            data["file"] = active_file["file"]
    return client._creoson_post("view", "activate", data)


def list_exploded(client, file_=None, name=None):
    """List views that match criteria and are exploded.

    Args:
        client (obj):
            creopyson Client.
        `file_` (str, optional):
            Model name. Defaults is current active model.
        name (str, optional):
            View name (wildcards allowed: True).
            Defaults is None: all views are listed.

    Returns:
        (list:str): List of view names.

    """
    data = {}
    if name is not None:
        data["name"] = name
    if file_ is not None:
        data["file"] = file_
    else:
        active_file = client.file_get_active()
        if active_file:
            data["file"] = active_file["file"]
    return client._creoson_post("view", "list_exploded", data, "viewlist")


def list_(client, file_=None, name=None):
    """List views that match criteria.

    Args:
        client (obj):
            creopyson Client.
        `file_` (str, optional):
            Model name. Defaults is current active model.
        name (str, optional):
            View name (wildcards allowed: True).
            Defaults is None: all views are listed.

    Returns:
        (list:str): List of view names.

    """
    data = {"name": "*"}
    if file_ is not None:
        data["file"] = file_
    else:
        active_file = client.file_get_active()
        if active_file:
            data["file"] = active_file["file"]
    if name is not None:
        data["name"] = name
    return client._creoson_post("view", "list", data, "viewlist")


def save(client, name, file_=None):
    """Save a model's current orientation as a new view.

    Args:
        client (obj):
            creopyson Client.
        name (str):
            View name.
        `file_` (str, optional):
            Model name. Defaults is current active model.

    Returns:
        None

    """
    data = {"name": name}
    if file_ is not None:
        data["file"] = file_
    else:
        active_file = client.file_get_active()
        if active_file:
            data["file"] = active_file["file"]
    return client._creoson_post("view", "save", data)
